@@ -32,9 +32,7 @@
 #
 # ==========================================================================
 
-# export USE_NEW_EDIRECT="true"
-
-version="14.6"
+version="16.6"
 
 # initialize common flags
 
@@ -43,6 +41,8 @@ dev=false
 internal=false
 external=false
 api_key=""
+immediate=false
+express=false
 
 email=""
 emailr=""
@@ -94,6 +94,39 @@ stp=0
 
 rest=""
 
+# set up colors for error report
+
+ColorSetup() {
+
+  if [ -z "$TERM" ] || [ ! -t 2 ]
+  then
+    RED=""
+    BLUE=""
+    BOLD=""
+    FLIP=""
+    INIT=""
+  elif command -v tput >/dev/null
+  then
+    RED="$(tput setaf 1)"
+    BLUE="$(tput setaf 4)"
+    BOLD="$(tput bold)"
+    FLIP="$(tput rev)"
+    INIT="$(tput sgr0)"
+  else
+    # assume ANSI
+    escape="$(printf '\033')"
+    RED="${escape}[31m"
+    BLUE="${escape}[34m"
+    BOLD="${escape}[1m"
+    FLIP="${escape}[7m"
+    INIT="${escape}[0m"
+  fi
+  LOUD="${INIT}${RED}${BOLD}"
+  INVT="${LOUD}${FLIP}"
+}
+
+ColorSetup
+
 # parse ENTREZ_DIRECT, eSearchResult, eLinkResult, or ePostResult
 
 ParseMessage() {
@@ -229,6 +262,16 @@ ParseCommonArgs() {
         argsConsumed=$(($argsConsumed + 1))
         shift
         ;;
+      -immediate )
+        immediate=true
+        argsConsumed=$(($argsConsumed + 1))
+        shift
+        ;;
+      -express )
+        express=true
+        argsConsumed=$(($argsConsumed + 1))
+        shift
+        ;;
       -base )
         argsConsumed=$(($argsConsumed + 1))
         shift
@@ -238,7 +281,7 @@ ParseCommonArgs() {
           argsConsumed=$(($argsConsumed + 1))
           shift
         else
-          echo "ERROR: Missing -base argument" >&2
+          echo "${INVT} ERROR: ${LOUD} Missing -base argument${INIT}" >&2
           exit 1
         fi
         ;;
@@ -251,7 +294,7 @@ ParseCommonArgs() {
           argsConsumed=$(($argsConsumed + 1))
           shift
         else
-          echo "ERROR: Missing -input argument" >&2
+          echo "${INVT} ERROR: ${LOUD} Missing -input argument${INIT}" >&2
           exit 1
         fi
         ;;
@@ -263,7 +306,7 @@ ParseCommonArgs() {
           web_env="$1"
           shift
         else
-          echo "ERROR: Missing -web argument" >&2
+          echo "${INVT} ERROR: ${LOUD} Missing -web argument${INIT}" >&2
           exit 1
         fi
         ;;
@@ -275,7 +318,7 @@ ParseCommonArgs() {
           stp="$1"
           shift
         else
-          echo "ERROR: Missing -step argument" >&2
+          echo "${INVT} ERROR: ${LOUD} Missing -step argument${INIT}" >&2
           exit 1
         fi
         ;;
@@ -288,7 +331,7 @@ ParseCommonArgs() {
           argsConsumed=$(($argsConsumed + 1))
           shift
         else
-          echo "ERROR: Missing -label argument" >&2
+          echo "${INVT} ERROR: ${LOUD} Missing -label argument${INIT}" >&2
           exit 1
         fi
         ;;
@@ -301,7 +344,7 @@ ParseCommonArgs() {
           argsConsumed=$(($argsConsumed + 1))
           shift
         else
-          echo "ERROR: Missing -email argument" >&2
+          echo "${INVT} ERROR: ${LOUD} Missing -email argument${INIT}" >&2
           exit 1
         fi
         ;;
@@ -314,7 +357,7 @@ ParseCommonArgs() {
           argsConsumed=$(($argsConsumed + 1))
           shift
         else
-          echo "ERROR: Missing -tool argument" >&2
+          echo "${INVT} ERROR: ${LOUD} Missing -tool argument${INIT}" >&2
           exit 1
         fi
         ;;
@@ -524,35 +567,14 @@ PrintQuery() {
 
 # three attempts for EUtils requests
 
-ColorSetup() {
-
-  if [ -z "$TERM" ] || [ ! -t 2 ]
-  then
-    red_fg=""
-    blue_fg=""
-    orig_colors=""
-  elif command -v tput >/dev/null
-  then
-    red_fg="$(tput setaf 1)"
-    blue_fg="$(tput setaf 4)"
-    orig_colors="$(tput op)"
-  else
-    # assume ANSI
-    escape="$(printf '\033')"
-    red_fg="$escape[31m"
-    blue_fg="$escape[34m"
-    orig_colors="$escape[39;49m"
-  fi
-}
-
 ErrorHead() {
 
   wrn="$1"
   whn="$2"
 
-  printf "${red_fg}${wrn}: FAILURE ( $whn )${orig_colors}\n" >&2
+  printf "${INVT} ${wrn}: ${LOUD} FAILURE ( $whn )${INIT}\n" >&2
   # display original command in blue letters
-  printf "${blue_fg}" >&2
+  printf "${BLUE}" >&2
 }
 
 ErrorTail() {
@@ -560,15 +582,16 @@ ErrorTail() {
   msg="$1"
   whc="$2"
 
-  printf "${orig_colors}" >&2
+  printf "${INIT}" >&2
   # display reformatted result in red letters
-  printf "${red_fg}${msg}${orig_colors}\n" >&2
+  printf "${RED}${msg}${INIT}\n" >&2
   if [ "$goOn" = true ]
   then
-    printf "${blue_fg}${whc} ATTEMPT${orig_colors}\n" >&2
+    printf "${BLUE}${whc} ATTEMPT" >&2
   else
-    printf "${blue_fg}QUERY FAILURE${orig_colors}\n" >&2
+    printf "${BLUE}QUERY FAILURE" >&2
   fi
+  printf "${INIT}\n" >&2
 }
 
 RequestWithRetry() {
@@ -580,8 +603,6 @@ RequestWithRetry() {
   # execute query
   res=$( "$@" )
 
-  ColorSetup
-
   warn="WARNING"
   whch="SECOND"
   while [ "$goOn" = true ]
@@ -606,7 +627,7 @@ RequestWithRetry() {
       *\<eFetchResult\>* | *\<eSummaryResult\>*  | *\<eSearchResult\>*  | *\<eLinkResult\>* | *\<ePostResult\>* | *\<eInfoResult\>* )
         case "$res" in
           *\<ERROR\>* )
-            ref=$( echo "$res" | xtract -format -doctype "" )
+            ref=$( echo "$res" | transmute -format indent -doctype "" )
             ErrorHead "$warn" "$when"
             PrintQuery "$@"
             if [ "$goOn" = true ]
@@ -621,7 +642,7 @@ RequestWithRetry() {
             res=$( "$@" )
             ;;
           *\<error\>* )
-            ref=$( echo "$res" | xtract -format -doctype "" )
+            ref=$( echo "$res" | transmute -format indent -doctype "" )
             ErrorHead "$warn" "$when"
             PrintQuery "$@"
             if [ "$goOn" = true ]
@@ -636,7 +657,7 @@ RequestWithRetry() {
             res=$( "$@" )
             ;;
           *\<ErrorList\>* )
-            ref=$( echo "$res" | xtract -format -doctype "" )
+            ref=$( echo "$res" | transmute -format indent -doctype "" )
             # question mark prints names of heterogeneous child objects
             errs=$( echo "$res" | xtract -pattern "ErrorList/*" -element "?" )
             if [ -n "$errs" ] && [ "$errs" = "PhraseNotFound" ]
@@ -659,7 +680,7 @@ RequestWithRetry() {
             fi
             ;;
           *\"error\":* )
-            ref=$( echo "$res" | xtract -format -doctype "" )
+            ref=$( echo "$res" | transmute -format indent -doctype "" )
             ErrorHead "$warn" "$when"
             PrintQuery "$@"
             ErrorTail "$ref" "$whch"
@@ -796,13 +817,18 @@ GenerateUidList() {
 
   if [ "$needHistory" = true ]
   then
-    GenerateHistoryChunks 25000 0 0 |
+    dbsx="$1"
+    chunk=25000
+    if [ "$dbsx" = "pubmed" ]
+    then
+      chunk=10000
+    fi
+    GenerateHistoryChunks "$chunk" 0 0 |
     while read fr chnk
     do
-      RunWithCommonArgs nquire -get "$base" esearch.fcgi \
+      RunWithCommonArgs nquire -get "$base" efetch.fcgi \
         -query_key "$qry_key" -WebEnv "$web_env" -retstart "$fr" -retmax "$chnk" \
-        -db "$1" -rettype uilist -retmode text |
-      xtract -pattern Id -element Id
+        -db "$dbsx" -rettype uilist -retmode text
     done
   else
     # otherwise obtain raw UIDs from available sources
@@ -818,7 +844,8 @@ GenerateUidList() {
     fi |
     # accn-at-a-time without case transformation
     sed 's/[^a-zA-Z0-9_.]/ /g; s/^ *//' |
-    fmt -w 1
+    fmt -w 1 |
+    grep '.'
   fi
 }
 
@@ -856,6 +883,76 @@ ExtractAccns() {
   fi
 }
 
+ExtractPMCIds() {
+
+  if [ "$needHistory" = false ]
+  then
+    GenerateUidList "$dbase" |
+    while read uid
+    do
+      case "$uid" in
+        PMC* )
+          echo "$uid" | sed -e 's/^PMC//g'
+          ;;
+        pmc* )
+          echo "$uid" | sed -e 's/^pmc//g'
+          ;;
+        * )
+          echo "$uid"
+          ;;
+      esac
+    done
+  fi
+}
+
+ExtractNucUids() {
+
+  # argument value: 1 = PACC, 2 = ACCN, 3 = integer and accession
+  kind="$1"
+
+  while read uid
+  do
+    case "$uid" in
+      *00000000 )
+        notInteger=$( echo "$uid" | sed -e 's/[0-9.]//g' )
+        if [ -n "$notInteger" ]
+        then
+          if [ "$kind" -eq 1 ]
+          then
+            echo "$uid"
+          fi
+        else
+          if [ "$kind" -eq 3 ]
+          then
+            echo "$uid"
+          fi
+        fi
+        ;;
+      *0000000 )
+        notInteger=$( echo "$uid" | sed -e 's/[0-9.]//g' )
+        if [ -n "$notInteger" ]
+        then
+          if [ "$kind" -eq 2 ]
+          then
+            echo "$uid"
+          fi
+        else
+          if [ "$kind" -eq 3 ]
+          then
+            echo "$uid"
+          fi
+        fi
+        ;;
+      * )
+        if [ "$kind" -eq 3 ]
+        then
+          echo "$uid"
+        fi
+        ;;
+    esac
+  done
+}
+
 ExtractPDB() {
 
   if [ "$needHistory" = false ]
@@ -863,18 +960,18 @@ ExtractPDB() {
     GenerateUidList "$dbase" |
     while read uid
     do
-    case "$uid" in
-      [0-9][0-9][0-9][0-9] )
-        # Four-digit UID
-        # peel off first to avoid mistaking for a chainless PDB ID
-        ;;
-      [0-9][0-9A-Za-z][0-9A-Za-z][0-9A-Za-z] | \
-      [0-9][0-9A-Za-z][0-9A-Za-z][0-9A-Za-z]_[A-Za-z]* )
-        # PDB ID
-        # properly case-sensitive only when untagged
-        echo "$uid"
-        ;;
-    esac
+      case "$uid" in
+        [0-9][0-9][0-9][0-9] )
+          # Four-digit UID
+          # peel off first to avoid mistaking for a chainless PDB ID
+          ;;
+        [0-9][0-9A-Za-z][0-9A-Za-z][0-9A-Za-z] | \
+        [0-9][0-9A-Za-z][0-9A-Za-z][0-9A-Za-z]_[A-Za-z]* )
+          # PDB ID
+          # properly case-sensitive only when untagged
+          echo "$uid"
+          ;;
+      esac
     done
   fi
 }
@@ -886,24 +983,24 @@ ExtractNonPDB() {
     GenerateUidList "$dbase" |
     while read uid
     do
-    case "$uid" in
-      [0-9][0-9][0-9][0-9] )
-        # Four-digit UID
-        # peel off first to avoid mistaking for a chainless PDB ID
-        echo "$uid"
-        ;;
-      [0-9][0-9A-Za-z][0-9A-Za-z][0-9A-Za-z] | \
-      [0-9][0-9A-Za-z][0-9A-Za-z][0-9A-Za-z]_[A-Za-z]* )
-        # PDB ID, skip
-        ;;
-      *[A-Za-z]* )
-        # accessions are already handled
-        echo "$uid"
-        ;;
-      * )
-        echo "$uid"
-        ;;
-    esac
+      case "$uid" in
+        [0-9][0-9][0-9][0-9] )
+          # Four-digit UID
+          # peel off first to avoid mistaking for a chainless PDB ID
+          echo "$uid"
+          ;;
+        [0-9][0-9A-Za-z][0-9A-Za-z][0-9A-Za-z] | \
+        [0-9][0-9A-Za-z][0-9A-Za-z][0-9A-Za-z]_[A-Za-z]* )
+          # PDB ID, skip
+          ;;
+        *[A-Za-z]* )
+          # accessions are already handled
+          echo "$uid"
+          ;;
+        * )
+          echo "$uid"
+          ;;
+      esac
     done
   fi
 }
@@ -914,11 +1011,21 @@ PrepareAccnQuery() {
   do
     echo "$uid+[$1]"
   done |
-  join-into-groups-of 10000 |
+  join-into-groups-of "$2" |
   sed -e 's/,/ OR /g' |
   tr '+' ' '
 }
 
+RunAccnSearch() {
+
+  while read qry
+  do
+    nquire -get "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi" \
+      -db "$dbase" -term "$qry" -retmax "$1" < /dev/null |
+    xtract -pattern eSearchResult -sep "\n" -element IdList/Id
+  done
+}
+
 PreparePDBQuery() {
 
   while read uid
@@ -929,18 +1036,36 @@ PreparePDBQuery() {
   sed -e 's/,/ OR /g'
 }
 
+PrepareSnpQuery() {
+
+  while read uid
+  do
+    case "$uid" in
+      rs* )
+        echo "$uid+[RS]"
+        ;;
+      ss* )
+        echo "$uid+[SS]"
+        ;;
+    esac
+  done |
+  join-into-groups-of 10000 |
+  sed -e 's/,/ OR /g' |
+  tr '+' ' '
+}
+
 LookupSpecialAccessions() {
 
   if [ "$needHistory" = false ]
   then
     fld=""
     case "$dbase" in
-      assembly | biosample | biosystems | cdd | dbvar | ipg | medgen | proteinclusters | seqannot | sra )
-        fld="ACCN"
-        ;;
-      annotinfo )
+      assembly | annotinfo )
         fld="ASAC"
         ;;
+      biosample | biosystems | cdd | dbvar | ipg | medgen | proteinclusters | seqannot | sra )
+        fld="ACCN"
+        ;;
       bioproject | genome )
         fld="PRJA"
         ;;
@@ -965,6 +1090,37 @@ LookupSpecialAccessions() {
       pcsubstance )
         fld="SRID"
         ;;
+      pmc* )
+        ids=$( ExtractPMCIds | fmt -w 1 | sort -n | uniq )
+        ;;
+      nuc* )
+        nucUidList=$( GenerateUidList "$dbase" )
+        anyNonInteger=$( echo "$nucUidList" | sed -e 's/[0-9.]//g' )
+        if [ -n "$anyNonInteger" ]
+        then
+          pacc=$( echo "$nucUidList" | ExtractNucUids "1" )
+          accn=$( echo "$nucUidList" | ExtractNucUids "2" )
+          lcl=$( echo "$nucUidList" | ExtractNucUids "3" )
+          pacres=""
+          accres=""
+          if [ -n "$pacc" ]
+          then
+            pacres=$( echo "$pacc" |
+                      PrepareAccnQuery "PACC" "100" |
+                      RunAccnSearch "1000" )
+          fi
+          if [ -n "$accn" ]
+          then
+            accres=$( echo "$accn" |
+                      PrepareAccnQuery "ACCN" "100" |
+                      RunAccnSearch "1000" )
+          fi
+          if [ -n "$pacres" ] || [ -n "$accres" ]
+          then
+            ids=$( echo "$pacres $accres $lcl" | fmt -w 1 | sort -n | uniq )
+          fi
+        fi
+        ;;
       protein )
         acc=$( ExtractPDB )
         lcl=$( ExtractNonPDB )
@@ -975,11 +1131,21 @@ LookupSpecialAccessions() {
           ids=$( echo "$rem $lcl" | fmt -w 1 | sort -n | uniq )
         fi
         ;;
+      snp )
+        acc=$( ExtractAccns )
+        lcl=$( ExtractUIDs )
+        if [ -n "$acc" ]
+        then
+          query=$( echo "$acc" | PrepareSnpQuery "$fld" )
+          rem=$( esearch -db "$dbase" -query "$query" | efetch -format uid )
+          ids=$( echo "$rem $lcl" | fmt -w 1 | sort -n | uniq )
+        fi
+        ;;
       taxonomy )
         acc=$( ExtractAccns )
         if [ -n "$acc" ]
         then
-          echo "ERROR: Taxonomy database does not index sequence accession numbers" >&2
+          echo "${INVT} ERROR: ${LOUD} Taxonomy database does not index sequence accession numbers${INIT}" >&2
           exit 1
         fi
         ;;
@@ -990,9 +1156,13 @@ LookupSpecialAccessions() {
       lcl=$( ExtractUIDs )
       if [ -n "$acc" ]
       then
-        query=$( echo "$acc" | PrepareAccnQuery "$fld" )
-        rem=$( esearch -db "$dbase" -query "$query" | efetch -format uid )
-        ids=$( echo "$rem $lcl" | fmt -w 1 | sort -n | uniq )
+        newids=$( echo "$acc" |
+                  PrepareAccnQuery "$fld" "1000" |
+                  RunAccnSearch "10000" )
+        if [ -n "$newids" ]
+        then
+          ids=$( echo "$newids $lcl" | fmt -w 1 | sort -n | uniq )
+        fi
       fi
     fi
   fi
